/*
File: HUFF2000.CPP
Description: Huffman Compression Implementation
Version: 1.0f
Copyright 2000, Ian Fernandes
File Format:	Byte0	1	2	3	4567				       8........           ...
					I	C	A	H   SizeofCompressedDatainbits NoOfSymbols-1	List Of Symbols 	Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream.h>
#include "BitIO16.Cpp"
#include <io.h>
#include <conio.h>
#include <fcntl.h>
#include <dir.h>
#include <ctype.h>

typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef unsigned char BYTE;

UINT NoOfSymbols;
ULONG TotalCount,compbitcount;

struct SYM_TABLE
{
	UINT Symbol;
	short CodeLen;
	UINT Code;
	float Freq;
	ULONG Count;
}  SymTable[256];

struct TreeData
{
	int *SymID;		//Symbol  ID - No In SymTable
	int No;
	float Freq;
	TreeData *up, *left, *right;
	TreeData *prev, *next;
}*root, *leaf, *stleaf;



char *Input;
char *Output;

FILE *hfInput, *hfOutput;

void PrintUsage()
{
	cout <<"\nHuffman Algorithm based Compression-Decompression Agent";
	cout <<"\nCopyright 2000, Ian Fernandes";
	cout <<"\nUsage:huff inputfile outputfile";
}
void CountSymbols()
{
	unsigned int bitread;
	unsigned char cbuff;
	for(int cntr=0;cntr<=255;cntr++)
	{
		SymTable[cntr].Count=0;
		SymTable[cntr].Symbol =cntr;
	}
//	BitIO inp;
	fseek(hfInput, 0 , SEEK_SET);
/*	if(inp.Start(hfInput, (BOOL)0, (BOOL)1))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}*/
	cout <<"\nCounting Symbols[                    ]\r";
	cout <<"Counting Symbols[";

	long charsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	charsread=0;
//	while(inp.StreamIn(&bitread	,8) != -1)
	while(!feof(hfInput))
	{
//		cbuff = (BYTE) bitread;
		cbuff = (BYTE) fgetc(hfInput);
		if (feof(hfInput))
			continue;

		++SymTable[cbuff].Count;
		++charsread;
		if(charsread > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}
	}
// 	inp.End();
	TotalCount=0;
	NoOfSymbols=0;
//	printf("\nBefore Sorting\n");
	for(cntr = 0 ;cntr<256;cntr++)
	{
		if(SymTable[cntr].Count !=0)
		{
//			printf("%c %i\t",cntr, SymTable[cntr].Count);
			TotalCount +=SymTable[cntr].Count;
			++NoOfSymbols;
		}
	}
	printf("\nTotal Count %.0f", (float)TotalCount);
	printf("\nTotal Symbols %i", NoOfSymbols);
}
int SortFunc(const void *a, const void *b)
{
	SYM_TABLE *d=(SYM_TABLE*)a;
	SYM_TABLE *e=(SYM_TABLE*)b;
	long result = d->Count - e->Count;
//	long result =((*(struct SYM_TABLE**)a)->Count - (*(struct SYM_TABLE**)b)->Count);
	if(result < 0L)
			return 1;
	if(result > 0L)
		return -1;
	return 0;
/*		return 1;
	if(((*(struct SYM_TABLE**)a)->Count) > ((*(struct SYM_TABLE**)b)->Count))
		return( -1);
	return 0;*/
}
void SortSymTable()
{
	qsort((void*)SymTable, (size_t)256,(size_t)sizeof(SymTable[0]),SortFunc);
/*	TotalCount=0;
	NoOfSymbols=0;
	printf("\nAfter Sorting\n");
	for(int cntr = 0 ;cntr<255;cntr++)
	{
		if(SymTable[cntr].Count !=0)
		{
			printf("%c %i\t",SymTable[cntr].Symbol, SymTable[cntr].Count);
			TotalCount +=SymTable[cntr].Count;
			++NoOfSymbols;
		}
	}
	printf("\nTotal Count  %.0f", (float)TotalCount);
	printf("\nTotal Symbols %i", NoOfSymbols);
*/
}

void GenFreqs()
{
	cout <<"\nGenerating freq...";
	for(int cntr =0; cntr <NoOfSymbols;cntr++)
	{
		SymTable[cntr].Freq = (float)SymTable[cntr].Count/(float)TotalCount;
/*		printf("\nChar %3i has count %6.0f with freq %10.8f",
				SymTable[cntr].Symbol,
				(float)SymTable[cntr].Count,
				SymTable[cntr].Freq);
		if(cntr %25 ==0)
			getch();*/
	}
	cout <<"Done";
}


void CreateLeaves()
{
	leaf = new (TreeData);
	stleaf = leaf;		//FIRST leaf is null leaf
	leaf->SymID = new int;
	leaf->SymID[0] = -1;
	leaf->Freq = 1.1;
	leaf->No = -1;
	leaf->up=leaf->left=leaf->right=leaf->prev=0;
	for(int cntr =0;cntr<NoOfSymbols;cntr++)
	{
		leaf->next =new TreeData;
		leaf->next->prev=leaf;
		leaf=leaf->next;
		leaf->SymID=new int;
		leaf->SymID[0] = cntr;
		leaf->Freq= SymTable[cntr].Freq;
		leaf->No=1;
		leaf->up =leaf->right=leaf->left=leaf->next=0;
	}
	leaf=stleaf;
/*	while(leaf != NULL)
	{
		printf("\nChar %3i has count %6.0f with freq %10.8f",
		SymTable[leaf->SymID[0]].Symbol,
		 (float)SymTable[leaf->SymID[0]].Count, leaf->Freq);
		leaf = leaf->next;
	}*/
}

Find_Low_Freq(TreeData **l1, TreeData **l2)
{
	TreeData *lf1,*lf2;
	TreeData *curl;
	lf1 = lf2 = stleaf;
	curl = stleaf;
	while(curl->next !=NULL)
	{
		curl = curl->next;
		if(curl->Freq <= lf1->Freq)
		{
			if(curl->Freq <= lf2->Freq)
			{
				lf1 = lf2;
				lf2 = curl;
			}
			else
				lf1 = curl;

		}
	}
	*l1=lf1;
	*l2=lf2;

}


void CreateTree()
{
	float freqcount =0;
	TreeData *lf1, *lf2, *uptree, *curl;
	int cntr;
	do
	{
		Find_Low_Freq(&lf1, &lf2);
		lf1->up = new TreeData;
		lf2->up = lf1->up;
		curl = lf1->up;
		curl->No = lf1->No +lf2->No;
		curl->Freq= lf1->Freq + lf2->Freq;
		freqcount = curl->Freq;
		curl->SymID = (int*)calloc(curl->No, sizeof(int));
		for(cntr =0; cntr < lf1->No ;cntr++)
			*(curl->SymID+cntr)=lf1->SymID[cntr];
		for(cntr =0; cntr < lf2->No ;cntr++)
			curl->SymID[cntr+lf1->No]=lf2->SymID[cntr];

		curl->left = lf1;
		curl->right = lf2;

		if(lf1->next == lf2)
		{
			lf1->prev->next = lf2->next;
			if(lf2->next !=NULL)
				lf2->next->prev = lf1->prev;
		}
		else if(lf2->next == lf1)
		{
			lf2->prev->next = lf1->next;
			if(lf1->next !=NULL)
				lf1->next->prev = lf2->prev;
		}

		else
		{
			lf1->prev->next = lf1->next;
			if(lf1->next !=NULL)
				lf1->next->prev = lf1->prev;
			lf2->prev->next = lf2->next;
//			if(lf2->next !=NULL)
				lf2->next->prev = lf2->prev;
		}
		leaf = stleaf;
		while(leaf->next != NULL)
			leaf = leaf->next;

		leaf->next  = curl;
		curl->next = 0;
		curl->prev = leaf;

/*		lf1->prev->next = curl;
		curl->prev=lf1->prev;
		curl->next = lf2->next;
		lf2->next->prev =curl;*/

//		printf("\nFreq of branch :%10.8f",freqcount);
		leaf = stleaf;
/*		while(leaf != NULL)
		{
			printf("\nChar %3i has count %6.0f with freq %10.8f",
			SymTable[leaf->SymID[0]].Symbol,
			 (float)SymTable[leaf->SymID[0]].Count, leaf->Freq);
			leaf = leaf->next;
		}
  */
/*		cout <<"\n------";
		leaf=leaf->prev;
		while(leaf->prev != NULL)
		{
			printf("\nChar %3i has count %6.0f with freq %10.8f",
			SymTable[leaf->SymID[0]].Symbol,
			 (float)SymTable[leaf->SymID[0]].Count, leaf->Freq);
			leaf = leaf->prev;
		}
*/
	}
	while(freqcount < 0.999999);

	root = curl;
}

void PrintTree(TreeData *temp, char node)
{
		if (temp ==NULL) return;

		printf("%c( %0.6f %i ",node,temp->Freq,temp->No);
//        if(temp->No ==1)
		for(int cntr =0;cntr <temp->No;cntr++)
			printf("%i ",temp->SymID[cntr]);
		printf(")\t");

		PrintTree(temp->left, 'L');
		PrintTree(temp->right, 'R');

}

void EncodeSym(TreeData* temp, UINT sym)
{

	UINT cntr=0;
	for(cntr=0;cntr<temp->left->No;cntr++)
	{
		if(temp->left->SymID[cntr] == sym)
		{
			SymTable[sym].Code =SymTable[sym].Code<<1;
			SymTable[sym].Code= SymTable[sym].Code | 0x0001;
			++SymTable[sym].CodeLen;
			EncodeSym(temp->left, sym);
			return;
		}
	}
	for(cntr=0;cntr<temp->right->No;cntr++)
	{
		if(temp->right->SymID[cntr] == sym)
		{
			SymTable[sym].Code =SymTable[sym].Code<<1;
			SymTable[sym].Code= SymTable[sym].Code | 0x0000;
			++SymTable[sym].CodeLen;
			EncodeSym(temp->right,sym);
			return;

		}
	}
}


void GenerateCode()
{
	for(int cntr=0; cntr<NoOfSymbols;cntr++)
	{
		SymTable[cntr].Code=0;
		SymTable[cntr].CodeLen=0;
		EncodeSym(root, cntr);
	}
}

void PrintCode()
{
	char bitseq[16];
	int sym, codelen;
	UINT code,tcode;
	float count;
	for(int cntr=0;cntr<NoOfSymbols;cntr++)
	{

		sym=SymTable[cntr].Symbol;
		count=(float)SymTable[cntr].Count;
		tcode=code=SymTable[cntr].Code;
		codelen=SymTable[cntr].CodeLen;
		for(int btcntr=0;btcntr<16;btcntr++)
			bitseq[btcntr]=0;
		for(btcntr=0;btcntr<codelen;btcntr++)
		{
			if(tcode&0x0001)
			 bitseq[codelen-btcntr-1]='1';
			else
			 bitseq[codelen-btcntr-1]='0';

			tcode = tcode>>1;
		}
		bitseq[btcntr]='\0';
		printf("\nSym-%3i Occ-%6.0f Code-%6i CodeBin-%16s len-%2i",
			sym,
				count,
					code,
						bitseq,
							codelen);

		if(cntr%25==0);
//			getch();
	}

}

void GenStats()
{
	compbitcount=0;
	ULONG codetablesize=0;
	ULONG csize;
	for(int cntr =0;cntr <NoOfSymbols;cntr++)
	{	compbitcount += SymTable[cntr].CodeLen*SymTable[cntr].Count;
		codetablesize += 8+SymTable[cntr].CodeLen+8;
	}
	printf("\nSize of Input file             %15.2f bytes",(float)TotalCount);
	printf("\nApprox size of Code table      %15.2f bytes",(float)(codetablesize/8.0));
	printf("\nApprox size of Data Code       %15.2f bits",(float)compbitcount);
	printf("\nApprox size of Compressed file %15.2f bytes",(float)(csize=((float)compbitcount+(float)codetablesize)/8.0));
	printf("\nCompression percent            %15.2f %",(float)(((float)TotalCount-(float)csize)/(float)TotalCount*100));
}
void WriteToFile()
{
	unsigned int ibuff;
	fseek(hfInput,0,SEEK_SET);
	fseek(hfOutput,0,SEEK_SET);
	fprintf(hfOutput,"ICAH") ;	//print the file header
	fwrite(&compbitcount,sizeof(ULONG),1,hfOutput);
	fprintf(hfOutput,"%c",(BYTE)(NoOfSymbols-1));
	BitIO outf;
	if(outf.Start(hfOutput,(BOOL)1,(BOOL)1))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	for(int cntr=0;cntr <NoOfSymbols;cntr++)
	{
		outf.StreamOut(SymTable[cntr].Symbol,8);
		outf.StreamOut(SymTable[cntr].CodeLen,8);
		outf.StreamOut(SymTable[cntr].Code,SymTable[cntr].CodeLen);
	}

	cout <<"\nEncoding file[                    ]\r";
	cout <<"Encoding file[";
	long charsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	charsread=0;

	while(!feof(hfInput))
	{
		ibuff=fgetc(hfInput);
		if(feof(hfInput))
			continue;
		cntr=0;
		while(SymTable[cntr].Symbol != ibuff)
		{
			cntr++;
			if(cntr>NoOfSymbols)
			{
				printf("\nLogical error");
				exit(1);
			}
		}
		outf.StreamOut(SymTable[cntr].Code,SymTable[cntr].CodeLen);
		++charsread;
		if(charsread > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}

	}

	outf.End();
}

int main (int argc, char **argv)
{

	clock_t tstart, tend;
	tstart=clock();
	PrintUsage();
	if( argc < 3)
	{
		cout <<"\nInsufficient Args";
		exit(1);
	}

	Input = new char[sizeof(argv[1])];
	Input = argv[1];
	if( argc >2)
	{
		Output =new char[sizeof(argv[2])];
		Output = argv[2];
	}

//Open File On Success returns 1 on failure exits with code 1
	if ((hfInput=fopen(Input,"rb")) == NULL)
	{
		cout <<"\n";
		perror(Input);
		exit(1);
	}
//	if(_DEBUG_)
		cout <<"\n"<<Input<<" opened successfully";

	struct ffblk filep;
	if(findfirst(Output,&filep,0))
	{

		cout <<"\n"<<Output<<" file exists. Overwrite(y/n)";
		if(tolower(getche()) == 'n')
		{	fclose(hfInput);
			exit(1);
		}
	}

	if((hfOutput=fopen(Output,"w+b")) ==NULL)
	{
		fclose(hfInput);
		perror (Output);
		exit(1);
	}

//	if(_DEBUG_)
//		cout <<"\n"<<Output<<" opened successfully";


	CountSymbols();
	SortSymTable();
	GenFreqs();
	CreateLeaves();
	CreateTree();
//	printf("\n");
//	PrintTree(root,'O');
	GenerateCode();
//	PrintCode();
	GenStats();
	WriteToFile();
	fclose(hfInput);
	fclose(hfOutput);

	tend=clock();
	cout <<"\nTime:"<<(tend-tstart)/CLK_TCK<<"\n";
	return 0;

}