/*
File: UNHUF.CPP
Version: 1.0f
Description: Huffman Compression Implementation
Copyright 2000, Ian Fernandes
File Format:	Byte0	1	2	3	4				5........           ...
					I	C	A	H   NoOfSymbols-1	List Of Symbols 	Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream.h>
#include "BitIO16A.Cpp"
#include <io.h>
#include <conio.h>
#include <fcntl.h>
#include <dir.h>
#include <ctype.h>
#include <string.h>

typedef unsigned int UINT;
typedef unsigned long ULONG;
typedef unsigned char BYTE;

UINT NoOfSymbols;
ULONG TotalCount;
ULONG CompSize;
FILE *hfInput, *hfOutput;
char *Input;
char *Output;
int CodeLenMax,CodeLenMin;
BitIO Hinpbits;

struct SYM_TABLE
{
	UINT Symbol;
	short CodeLen;
	UINT Code;
	float Freq;
	ULONG Count;
}  SymTable[256];

void PrintUsage()
{
	cout <<"\nHuffman Algorithm based Compression-Decompression Agent";
	cout <<"\nCopyright 2000, Ian Fernandes";
	cout <<"\nUsage:unhuff inputfile outputfile";
}

VerifyHeader()
{
	unsigned char Hdr[5];
	fseek(hfInput, 0, SEEK_SET);
	if(fread(Hdr, 1, 4, hfInput) !=4)
	{
		cout<<"\nUnable to read header";
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	Hdr[4]='\0';
	if( strcmp(Hdr,"ICAH") !=0)
	{
		cout <<"\nNot ICAH archive";
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	cout <<"\nHeader is "<<Hdr<<" ";
	fread(&CompSize,sizeof(ULONG),1,hfInput);
	printf("\nBits in compressed field %10.2f",(float)CompSize);
	NoOfSymbols=fgetc(hfInput)+1;
	printf("\nNo of Chars is %i", NoOfSymbols);

}

void CreateDecodeTable()
{
	UINT inbits;
	if(Hinpbits.Start(hfInput, (BOOL)0, (BOOL)1))
	{
		fclose(hfInput);
		fclose(hfOutput);
		exit(1);
	}
	CodeLenMax=0;
	CodeLenMin=32766;
	for (int cntr=0;cntr<NoOfSymbols;cntr++)
	{
		if(Hinpbits.StreamIn(&inbits, 8)==-1)
		{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}

		SymTable[cntr].Symbol=inbits;
		if(Hinpbits.StreamIn(&inbits, 8)==-1)
		{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}

		SymTable[cntr].CodeLen=inbits;
		if(SymTable[cntr].CodeLen> CodeLenMax)
			CodeLenMax=SymTable[cntr].CodeLen;
		if(SymTable[cntr].CodeLen< CodeLenMin)
			CodeLenMin=SymTable[cntr].CodeLen;
		if(Hinpbits.StreamIn(&inbits, SymTable[cntr].CodeLen)==-1)
		{
			fclose(hfInput);
			fclose(hfOutput);
			exit(1);
		}

		SymTable[cntr].Code= inbits;
	}
}

void PrintCode()
{
	char bitseq[16];
	int sym, codelen;
	UINT code,tcode;
	float count;
	for(int cntr=0;cntr<NoOfSymbols;cntr++)
	{

		sym=SymTable[cntr].Symbol;
		count=(float)SymTable[cntr].Count;
		tcode=code=SymTable[cntr].Code;
		codelen=SymTable[cntr].CodeLen;
		for(int btcntr=0;btcntr<16;btcntr++)
			bitseq[btcntr]=0;
		for(btcntr=0;btcntr<codelen;btcntr++)
		{
			if(tcode&0x0001)
			 bitseq[codelen-btcntr-1]='1';
			else
			 bitseq[codelen-btcntr-1]='0';

			tcode = tcode>>1;
		}
		bitseq[btcntr]='\0';
/*		printf("\nSym-%3i Occ-%6.0f Code- %6u CodeBin-%16s len-%2i",
			sym,
				count,
					ShiftToMSB(code,codelen),
						bitseq,
							codelen);
  */
		if(cntr%25==0);
//			getch();
	}

}
void DecodeFile()
{
	BitIO outbytes;
	fseek(hfOutput,0,SEEK_SET);
	int bitstoread=CodeLenMax;
	int actualbitsread,actuallen;
	int cntr, bitcntr,templen;
	UINT bitsread, tempbits,tempcode, inpbuff;
	BOOL codefound;
	tempbits=0;
//	if(outbytes.Start(hfOutput, (BOOL)1, (BOOL)1))
	{
  //		fclose(hfInput);
	//	fclose(hfOutput);
	  //	exit(1);
	}
	inpbuff =0;
	cout <<"\nDecoding data[                    ]\r";
	cout <<"Decoding data[";

	long charsread, mult;
	int handle;
	long filelen= filelength(handle = (open(Input,O_RDONLY)));
	close (handle);
	mult =0;
	charsread=0;
	BOOL getfromstream=TRUE;
	BOOL donedecoding=FALSE;
	ULONG totalbitsread=0;
	while(totalbitsread < CompSize)
	{
		actualbitsread=0;
		if(totalbitsread+bitcntr+bitstoread > CompSize)
			bitstoread = CompSize-totalbitsread-bitcntr;
		if(bitstoread<0)
			bitstoread = 0;

		actualbitsread=Hinpbits.StreamIn(&bitsread, bitstoread);
		if(actualbitsread == -1)
			exit(1);


		actuallen=bitcntr+actualbitsread;
		inpbuff =ShiftToLSB(inpbuff,actuallen);
		bitsread = inpbuff |bitsread;
		codefound=FALSE;
		bitcntr=0;
		while(!codefound)
		{	tempbits=bitsread>>bitcntr;
//			tempbits=ShiftToMSB(tempbits, CodeLenMax-bitcntr);
			tempbits=ShiftToMSB(tempbits, actuallen-bitcntr);
			for(cntr=0;cntr<NoOfSymbols;cntr++)
			{
				tempcode=ShiftToMSB(SymTable[cntr].Code,SymTable[cntr].CodeLen);
				templen =SymTable[cntr].CodeLen;
				if(tempcode == tempbits && templen == (actuallen-bitcntr))
				{
					codefound=TRUE;
					totalbitsread += SymTable[cntr].CodeLen;
					if(totalbitsread>CompSize)
						break;
					fputc(SymTable[cntr].Symbol, hfOutput);
//					outbytes.StreamOut(SymTable[cntr].Symbol,8);
					break;
				}
			}
			if(!codefound)
				++bitcntr;
			if(bitcntr>CodeLenMax)
			{
				printf("\nLogical Error");
				fclose(hfInput);
				fclose(hfOutput);
				exit(1);
			}
		}

		charsread+=actualbitsread;
		if(charsread/8 > filelen*mult/100)
		{
			mult+=5;
			cout <<".";
		}

		bitstoread= CodeLenMax-bitcntr;
		if(bitstoread ==CodeLenMax)
			inpbuff=0;
		else
			inpbuff = ShiftToMSB(bitsread,bitcntr);


	}


	Hinpbits.End();
//	outbytes.End();
}






int main (int argc, char **argv)
{

	clock_t tstart, tend;
	tstart=clock();
	PrintUsage();
	if( argc < 3)
	{
		cout <<"\nInsufficient Args";
		exit(1);
	}

	Input = new char[sizeof(argv[1])];
	Input = argv[1];
	if( argc >2)
	{
		Output =new char[sizeof(argv[2])];
		Output = argv[2];
	}

//Open File On Success returns 1 on failure exits with code 1
	if ((hfInput=fopen(Input,"rb")) == NULL)
	{
		cout <<"\n";
		perror(Input);
		exit(1);
	}
//	if(_DEBUG_)
		cout <<"\n"<<Input<<" opened successfully";
	struct ffblk filep;
	if(findfirst(Output,&filep,0))
	{
		cout <<"\n"<<Output<<" file exists. Overwrite(y/n)";
		if(tolower(getche()) == 'n')
		{	fclose(hfInput);
			exit(1);
		}
	}

	if((hfOutput=fopen(Output,"w+b")) ==NULL)
	{
		fclose(hfInput);
		perror (Output);
		exit(1);
	}

//	if(_DEBUG_)
//		cout <<"\n"<<Output<<" opened successfully";

	VerifyHeader();
	CreateDecodeTable();
	PrintCode();
	DecodeFile();
	tend=clock();
	cout <<"\nTime:"<<(tend-tstart)/CLK_TCK<<"\n";
	return 0;

}